/*
  File name: CMFile.cpp
  Copyright: (C) 2007 Tomasz Kazmierczak

  Creation date: 23.04.2007
  Last modification date: 22.08.2007

  This file is part of Cytadela

 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *

*/

#include "CMFile.h"
#include "File3Dg.h"

bool CMFile::open(const char *filename)
{
	//TODO: assert filename

	close();

	m3DgFilePath = filename;
	std::string cmfile = m3DgFilePath + ".cmf";

	mFilePtr = fopen(cmfile.c_str(), "rb");
	if(not mFilePtr)
		return false;

	return true;
}

void CMFile::close()
{
	if(mFilePtr)
		fclose(mFilePtr);
	mFilePtr = 0;
	m3DgFilePath.clear();
}

bool CMFile::load(GOInterfacePtr &objects)
{
	//TODO: assert objects
	//TODO: assert mFilePtr

	I3Dg *geometry = (I3Dg *)objects;
	if(not ((File3Dg *)geometry)->read(m3DgFilePath.c_str()))
		return false;

	//read the objects info and create game objects
	if(not readObjectsInfo(objects))
		return false;

	//map data
	uint32_t size;
	if(not readMapDataSize(size))
		return false;
	//alloc memory for the data
	MapDataPtr ptr = objects->allocMapData(size);
	if(not ptr)
		return false;
	//read the map data
	if(not readMapData(ptr))
		return false;

	return true;
}

bool CMFile::readMapDataSize(uint32_t &size)
{
	//TODO: assert mFilePtr

	fseek(mFilePtr, 2 * sizeof(Point3Dg), SEEK_SET);
	//read size of the data
	if(fread(&size, 1, sizeof(uint32_t), mFilePtr) < sizeof(uint32_t))
		return false;

	return true;
}

bool CMFile::readMapData(MapDataPtr &data)
{
	//TODO: assert mFilePtr

	fseek(mFilePtr, 2 * sizeof(Point3Dg), SEEK_SET);
	//read size of the data
	uint32_t size;
	if(fread(&size, 1, sizeof(uint32_t), mFilePtr) < sizeof(uint32_t))
		return false;
	//read the data
	if(fread(data, 1, size, mFilePtr) < size)
		return false;
	return true;
}

bool CMFile::readInitialPosition(Point3Dg &position)
{
	//TODO: assert mFilePtr

	fseek(mFilePtr, 0, SEEK_SET);
	if(fread(&position, 1, sizeof(Point3Dg), mFilePtr) < sizeof(Point3Dg))
		return false;
	return true;
}

bool CMFile::readInitialDirection(Point3Dg &direction)
{
	//TODO: assert mFilePtr

	fseek(mFilePtr, sizeof(Point3Dg), SEEK_SET);
	if(fread(&direction, 1, sizeof(Point3Dg), mFilePtr) < sizeof(Point3Dg))
		return false;
	return true;
}

bool CMFile::readObjectsInfo(GOInterfacePtr objects)
{
	//TODO: assert mFilePtr

	//skip non-objects' data
	fseek(mFilePtr, 2 * sizeof(Point3Dg), SEEK_SET);
	//read size of the data
	uint32_t size;
	if(fread(&size, 1, sizeof(uint32_t), mFilePtr) < sizeof(uint32_t))
		return false;
	fseek(mFilePtr, size, SEEK_CUR);

	//read the data
	uint8_t type;
	while(fread(&type, 1, sizeof(uint8_t), mFilePtr) == sizeof(uint8_t))
	{
		//UID
		Mesh3DgUID uid;
		if(fread(&uid, 1, sizeof(Mesh3DgUID), mFilePtr) < sizeof(Mesh3DgUID))
			return false;
		//player collision type
		uint8_t playerCollision;
		if(fread(&playerCollision, 1, sizeof(uint8_t), mFilePtr) < sizeof(uint8_t))
			return false;
		//ammo collision type
		uint8_t ammoCollision;
		if(fread(&ammoCollision, 1, sizeof(uint8_t), mFilePtr) < sizeof(uint8_t))
			return false;
		//type specific info
		switch(type)
		{
			case GOTYPE_WALL:
			{
				GOWall *wall = (GOWall *)objects->createGameObject(GOTYPE_WALL, objects->mMeshes[uid]);
				if(not wall)
					return false;
				wall->setPlayerCollisionType((PlayerCollisionType)playerCollision);
				wall->setAmmoCollisionType((AmmoCollisionType)ammoCollision);
				//read the rest of the data
				Mesh3DgUID slotMUID;
				uint16_t offset;
				fread(&slotMUID, 1, sizeof(Mesh3DgUID), mFilePtr);
				if(fread(&offset, 1, sizeof(uint16_t), mFilePtr) < sizeof(uint16_t))
					return false;
				//set the data
				if(slotMUID)
					wall->setSlotMesh(objects->mMeshes[slotMUID]);
				wall->setSlotActionDataOffset(offset);
			}
			break;
			case GOTYPE_SPRITE:
			{
				GOSprite *sprite = (GOSprite *)objects->createGameObject(GOTYPE_SPRITE, objects->mMeshes[uid]);
				if(not sprite)
					return false;
				int16_t hp;
				if(fread(&hp, 1, sizeof(int16_t), mFilePtr) < sizeof(int16_t))
					return false;
				sprite->setHP(hp);
				sprite->setPlayerCollisionType((PlayerCollisionType)playerCollision);
				sprite->setAmmoCollisionType((AmmoCollisionType)ammoCollision);
			}
			break;
			case GOTYPE_DOOR:
			{
				GODoor *door = (GODoor *)objects->createGameObject(GOTYPE_DOOR, objects->mMeshes[uid]);
				if(not door)
					return false;
				door->setPlayerCollisionType((PlayerCollisionType)playerCollision);
				door->setAmmoCollisionType((AmmoCollisionType)ammoCollision);
				//read the rest of the data
				Mesh3DgUID frameMUID;
				uint8_t doorType;
				fread(&frameMUID, 1, sizeof(Mesh3DgUID), mFilePtr);
				if(fread(&doorType, 1, sizeof(uint8_t), mFilePtr) < sizeof(uint8_t))
					return false;
				//set the data
				if(not frameMUID)
					return false;
				door->setFrameMesh(objects->mMeshes[frameMUID]);
				door->setDoorType((DoorType)doorType);
				//add the object to the interface
			}
			break;
		}
	}
	return true;
}
