/*
  File name: GOInterface.h
  Copyright: (C) 2007 Tomasz Kazmierczak

  Creation date: 23.04.2007
  Last modification date: 8.09.2007

  This file is part of Cytadela

 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *

*/

#ifndef _GO_INTERFACE_H_
#define _GO_INTERFACE_H_

#include "I3Dg.h"

typedef enum {
	GOTYPE_WALL,
	GOTYPE_SPRITE,
	GOTYPE_DOOR
} GameObjectType;

typedef enum {
	PCTYPE_NONE,
	PCTYPE_BUMP,
	PCTYPE_STOP,
	PCTYPE_TELEPORT
} PlayerCollisionType;

typedef enum {
	ACTYPE_NONE,
	ACTYPE_UNBREAKABLE,
	ACTYPE_EXPLODABLE
} AmmoCollisionType;

typedef enum {
	DOORTYPE_UNDEFINED,
	DOORTYPE_VERTICAL_1,
	DOORTYPE_VERTICAL_2,
	DOORTYPE_HORIZONTAL,
	DOORTYPE_NUM_OF_TYPES
} DoorType;

class GameObject : public Object3Dg {
protected:
	GameObjectType      mType;
	PlayerCollisionType mPlayerCollision;
	AmmoCollisionType   mAmmoCollision;

public:
	GameObject(GameObjectType type, Mesh3Dg *mesh) : mType(type), mPlayerCollision(PCTYPE_NONE),
						 mAmmoCollision(ACTYPE_NONE)
	{ setParent((Object3Dg *)mesh); mesh->setChild((Object3Dg *)this); };

	//set
	void setPlayerCollisionType(PlayerCollisionType type) { mPlayerCollision = type; };
	void setAmmoCollisionType(AmmoCollisionType type) { mAmmoCollision = type; };
	//get
	Mesh3Dg *getMesh() { return (Mesh3Dg *)getParent(); };
	GameObjectType getType() { return mType; };
	PlayerCollisionType getPlayerCollisionType() { return mPlayerCollision; };
	AmmoCollisionType getAmmoCollisionType() { return mAmmoCollision; };
};

class GOWall : public GameObject {
private:
	Mesh3Dg  *mSlotMesh;
	uint16_t  mSlotActionDataOffset;
public:
	GOWall(Mesh3Dg *mesh) : GameObject(GOTYPE_WALL, mesh), mSlotMesh(0) {};

	//set
	void setSlotMesh(Mesh3Dg *mesh) { mSlotMesh = mesh; };
	void setSlotActionDataOffset(uint16_t offset) { mSlotActionDataOffset = offset; };
	//get
	Mesh3Dg *getSlotMesh() { return mSlotMesh; };
	uint16_t getSlotActionDataOffset() { return mSlotActionDataOffset; };
};

class GODoor : public GameObject {
private:
	Mesh3Dg *mFrameMesh;
	DoorType mDoorType;
public:
	GODoor(Mesh3Dg *mesh) : GameObject(GOTYPE_DOOR, mesh), mFrameMesh(0) {};

	//set
	void setFrameMesh(Mesh3Dg *mesh) { mFrameMesh = mesh; };
	void setDoorType(DoorType type) { mDoorType = type; };
	//get
	Mesh3Dg *getFrameMesh() { return mFrameMesh; };
	DoorType getDoorType() { return mDoorType; };
};

class GOSprite : public GameObject {
private:
	int16_t mHP;
public:
	GOSprite(Mesh3Dg *mesh) : GameObject(GOTYPE_SPRITE, mesh), mHP(0) {};

	//set
	void setHP(int16_t hp) { mHP = hp; };
	//get
	int16_t getHP() { return mHP; };
	//other
	int16_t decreaseHP(int16_t points) { mHP -= points; return mHP; };
};

class TemporaryObject : public GOSprite {
private:
	uint32_t  mLifeTime;
	uint32_t  mTimeCounter;
	uint32_t  mAnimStartDelay;
	Vector3Dg mVelocityDir;
	float     mRange;
	float     mDistaince;
	void     *mUserData;

public:
	TemporaryObject(Mesh3Dg *mesh, uint32_t time) : GOSprite(mesh), mLifeTime(time),
					mTimeCounter(0), mAnimStartDelay(0), mVelocityDir(0.0f, 0.0f, 0.0f),
					mRange(0.0f), mDistaince(0.0f), mUserData(0) {};

	TemporaryObject(Mesh3Dg *mesh, float range, Vector3Dg direction) : GOSprite(mesh), mLifeTime(0),
					mTimeCounter(0), mAnimStartDelay(0), mVelocityDir(direction),
					mRange(range), mDistaince(0.0f), mUserData(0) {};

	bool update(uint32_t time);
	bool update(uint32_t time, float distaince);
	void setAnimStartDelay(uint32_t delay) { mAnimStartDelay = delay; };
	void setVelocityDirection(Vector3Dg &velocitydir) { mVelocityDir = velocitydir; };
	void setUserData(void *data) { mUserData = data; };
	Vector3Dg *getVelocityDirection() { return &mVelocityDir; };
	uint32_t getLifeTime() { return mLifeTime; };
	float getRange() { return mRange; };
	void *getUserData() { return mUserData; };
};

typedef std::map<Mesh3DgUID, GameObject *> GameObjectsMap;

class GOInterface : public I3Dg {
protected:
	GameObjectsMap  mGameObjects;
	uint32_t        mMapDataSize;
	uint16_t       *mMapData;

	bool addGameObject(GameObject *object);
	bool remGameObject(GameObject *object);
public:
	GOInterface() : mMapDataSize(0), mMapData(0) {};
	~GOInterface();
	void release() { delete this; };

	/// @brief creates game object of a given type and places it in the objects' map
	/// You should not directly delete objects created this way - use destroyGameObject() instead
	/// If you won't destroy an object created by this function, the destructor will do it for you
	GameObject *createGameObject(GameObjectType type, Mesh3Dg *mesh);
	/// @brief creates a temporary game object
	/// You should not directly delete objects created this way - use destroyGameObject() instead
	/// If you won't destroy an object created by this function, the destructor will do it for you
	TemporaryObject *createTempGameObject(Mesh3Dg *mesh, uint32_t time);
	TemporaryObject *createTempGameObject(Mesh3Dg *mesh, float range, Vector3Dg direction);
	/// @brief removes a given object from the objects' map and deletes it from the memory
	/// With this function you should only destroy objects created with createGameObject()
	bool destroyGameObject(GameObject *object);

	uint16_t *allocMapData(uint32_t size);
	//get
	uint16_t *getMapDataPtr() { return mMapData; };
	uint32_t getMapDataSize() { return mMapDataSize; };
	GameObject *getGameObject(Mesh3DgUID muid);
	GameObjectType getObjectType(Mesh3Dg *mesh);
	GameObjectsMap::iterator firstGameObject() { return mGameObjects.begin(); };
	GameObjectsMap::iterator lastGameObject() { return mGameObjects.end(); };
};

GOInterface *createGOInterface();

#endif //_GO_INTERFACE_H_
