/*
  File name: farread.cpp
  Copyright: (C) 2005 - 2007 Tomasz Kazmierczak

  Creation date: 27.07.2005
  Last modification date: 28.08.2007

  This file is part of Cytadela

 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *

*/

#include "farread.h"
#include <stdio.h>
#include <memory.h>

class TGACMInfo {
public:
	uint16_t  wCMFirstEntry;   //position of first color map entry
	uint16_t  wCMLength;       //how many entries in color map
	uint8_t  btCMEntrySize;   //size of one color map entry
};

class TGAProps {
public:
	uint16_t  wOriginX;        //image origin coords
	uint16_t  wOriginY;
	uint16_t  wWidth;
	uint16_t  wHeight;
	uint8_t  btBitsPerPixel;
	uint8_t  btImgDescriptor; //image desctiption offset
};

uint8_t *readTGAData(FILE *pFARFile, uint32_t &width, uint32_t &height, uint32_t &bpp, uint32_t &size)
{
	//id field length
	uint8_t  btIDFieldLength;
	if(not fread(&btIDFieldLength, 1, 1, pFARFile))
		return 0;

	uint8_t  btColorMapType;
	if(not fread(&btColorMapType, 1, 1, pFARFile))
		return 0;

	uint8_t  btImageType;
	if(not fread(&btImageType, 1, 1, pFARFile))
		return 0;

	//check image type
	switch(btImageType)
	{
	//if one of these - ok
	case 1:
	case 2:
	case 3:
	case 9:
	case 10:
	case 11:
		break;

	//if any other - we don't read
	default:
		{
			fprintf(stderr, "Error: unsupported TGA file type\n");
			return 0;
		}
	}

	TGACMInfo sCMInfo;
	if(fread(&sCMInfo, 1, 5, pFARFile) < 5)
		return 0;
	//check the size of a single color map entry
	if(btColorMapType)
	{
		switch(sCMInfo.btCMEntrySize)
		{
		case 15:
		case 16:
		case 24:
		case 32:
			break;

		default:
			{
				fprintf(stderr, "Error: unsupported TGA file type\n");
				return 0;
			}
		}
	}


	TGAProps sProps;
	if(fread(&sProps, 1, 10, pFARFile) < 10)
		return 0;
	//check the image color depth
	switch(sProps.btBitsPerPixel)
	{
	case 8:
	case 16:
	case 24:
	case 32:
		break;

	default:
		{
			fprintf(stderr, "Error: unsupported TGA file type\n");
			return 0;
		}
	}

	//skip the id filed
	fseek(pFARFile, btIDFieldLength, SEEK_CUR);

	//read color map if any
	uint8_t *pColorMap = 0;
	if(btColorMapType == 1)
	{
		//calculate color map size
		uint32_t dwCMSize = sCMInfo.wCMLength;
		if(sCMInfo.btCMEntrySize == 15)
			dwCMSize *= 2;
		else
			dwCMSize *= (sCMInfo.btCMEntrySize / 8);

		//mem for color map
		pColorMap = new uint8_t[dwCMSize];
		if(not pColorMap)
		{
			fprintf(stderr, "Error: out of memory\n");
			return 0;
		}

		//read the color map
		if(fread(pColorMap, 1, dwCMSize, pFARFile) < dwCMSize)
		{
			delete[] pColorMap;
			return 0;
		}
	}

	//data size (if the image is in grayscale then the bytes count is equal to the pixels count)
	uint32_t dwBitsSize = sProps.wWidth * sProps.wHeight;
	//if there is no color map, multiply the size by bytes per pixel count
	if((btImageType == 2) or (btImageType == 10))
		dwBitsSize *= (sProps.btBitsPerPixel / 8);
	//if there is the color map then we will convert the image to format
	//w/o color map, so we multiply the size by size of pixel in the color map
	if((btImageType == 1) or (btImageType == 9))
	{
		if(sCMInfo.btCMEntrySize == 15)
			dwBitsSize *= 2;
		else
			dwBitsSize *= (sCMInfo.btCMEntrySize / 8);
	}

	//memory for pixels
	uint8_t *pReadData = new uint8_t[dwBitsSize];
	if(not pReadData)
	{
		if(pColorMap)
			delete[] pColorMap;
		fprintf(stderr, "Error: out of memory\n");
		return 0;
	}

	//read the data
	switch(btImageType)
	{
	//uncompressed w/ color map
	case 1:
		{
			//place for pixel bytes
			uint32_t dwPixel = 0;

			//read pixels (pixel value is a position in color map)
			for(uint32_t i = 0; i < (uint32_t)(sProps.wWidth * sProps.wHeight); i++)
			{
				if(fread(&dwPixel, 1, (uint32_t)(sProps.btBitsPerPixel / 8), pFARFile)
									< (uint32_t)(sProps.btBitsPerPixel / 8))
				{
					delete[] pReadData;
					delete[] pColorMap;
					return 0;
				}

				//conversion to image w/o color map
				//(to image data write a color value from the color map)
				//calculate how many bytes to copy (bytes per pixel)
				uint32_t dwBytesCount = dwBitsSize / (sProps.wWidth * sProps.wHeight);

				for(uint32_t j = 0; j < dwBytesCount; j++)
					pReadData[(dwBytesCount * i) + j] = pColorMap[(dwBytesCount * dwPixel) + j];
			}
		}
		break;

	//uncompressed image w/o color map
	case 2:
	case 3:
		if(fread(pReadData, 1, dwBitsSize, pFARFile) < dwBitsSize)
		{
			delete[] pReadData;
			return 0;
		}
		break;

	//compressed (RLE) w/ color map
	case 9:
		{
			//place for a packet header
			uint8_t btPacketH;
			//place for a pixel
			uint32_t dwPixel = 0;
			//bytes per pixel count
			uint32_t dwBytesCount = dwBitsSize / (sProps.wWidth * sProps.wHeight);

			//bytes counter
			uint32_t i = 0;
			//read packets till all pixels are cpied to image data
			while(i < dwBitsSize)
			{
				//read the packet header
				if(not (fread(&btPacketH, 1, 1, pFARFile)))
				{
					delete[] pReadData;
					delete[] pColorMap;
					return 0;
				}

				//if RL packet
				if(btPacketH & 128)
				{
					//how many pixels in packet? (value of 7 LSBs + 1)
					btPacketH -= 127;

					//read pixel bytes
					if(fread(&dwPixel, 1, (uint32_t)(sProps.btBitsPerPixel / 8), pFARFile)
										< (uint32_t)(sProps.btBitsPerPixel / 8))
					{
						delete[] pReadData;
						delete[] pColorMap;
						return 0;
					}

					//copy data of pixel from color map to image data
					for(uint8_t j = 0; j < btPacketH; j++)
					{
						//copy every byte separately
						for(uint8_t k = 0; k < dwBytesCount; k++, i++)
							pReadData[i] = pColorMap[(dwBytesCount * dwPixel) + k];
					}
				}
				//if non RL packet
				else
				{
					//pixels in packet (value of 7 LSBs + 1)
					btPacketH++;

					//read pixels and convert to non color mapped image
					for(uint8_t j = 0; j < btPacketH; j++)
					{
						//read the place in color map rot the pixel
						if(fread(&dwPixel, 1, (sProps.btBitsPerPixel / 8), pFARFile)
											< (sProps.btBitsPerPixel / 8))
						{
							delete[] pReadData;
							delete[] pColorMap;
							return 0;
						}

						//copy bytes from color map to image data
						for(uint8_t k = 0; k < dwBytesCount; k++, i++)
							pReadData[i] = pColorMap[(dwBytesCount * dwPixel) + k];
					}
				}
			}
		}
		break;

	//compressed (RLE) w/o color map
	case 10:
	case 11:
		{
			//place for a packet header
			uint8_t btPacketH;

			//bytes counter
			uint32_t i = 0;
			//read packets till all bytes are copied
			while(i < dwBitsSize)
			{
				//read the packet header
				if(not (fread(&btPacketH, 1, 1, pFARFile)))
				{
					delete[] pReadData;
					return 0;
				}

				//if RL packet
				if(btPacketH & 128)
				{
					//how many pixels in packet? (value of 7 LSBs + 1)
					btPacketH -= 127;

					//place for a pixel
					uint8_t *btTmpPixel = new uint8_t[(sProps.btBitsPerPixel / 8)];
					if(not btTmpPixel)
					{
						delete[] pReadData;
						fprintf(stderr, "Error: out of memory\n");
						return 0;
					}

					//read pixel bytes
					if(fread(btTmpPixel, 1, (sProps.btBitsPerPixel / 8), pFARFile)
											< (sProps.btBitsPerPixel / 8))
					{
						delete[] btTmpPixel;
						delete[] pReadData;
						return 0;
					}

					//copy the pixel to image data
					for(uint8_t j = 0; j < btPacketH; j++)
					{
						//copy each byte separately
						for(uint8_t k = 0; k < (sProps.btBitsPerPixel / 8); k++, i++)
							pReadData[i] = btTmpPixel[k];
					}

					//delete temporary pixel memory
					delete[] btTmpPixel;
				}
				//if non RL packet
				else
				{
					//pixels in packet (value of 7 LSBs + 1)
					btPacketH++;

					//copy pixels to the image
					if(fread(&pReadData[i], 1, (btPacketH * (sProps.btBitsPerPixel / 8)), pFARFile) <
												(btPacketH * (sProps.btBitsPerPixel / 8)))
					{
						delete[] pReadData;
						return 0;
					}

					//increase bytes counter
					i += (btPacketH * (sProps.btBitsPerPixel / 8));
				}
			}
		}
		break;
	}

	//if we used color map than we must free it
	if(pColorMap)
		delete[] pColorMap;

	//flip the image if need to
	//flip up <-> down
	if(sProps.btImgDescriptor & 32)
	{
		uint8_t *tmp = new uint8_t[dwBitsSize];
		if(not tmp)
		{
			delete[] pReadData;
			fprintf(stderr, "Error: out of memory\n");
			return 0;
		}

		//copy the image to temp memory
		memcpy(tmp, pReadData, dwBitsSize);

		//copy lines back to image memory in reversed order
		for(uint32_t i = 0; i < sProps.wHeight; i++)
		{
			//copy each line separately
			uint32_t linewidth = dwBitsSize / sProps.wHeight;
			uint32_t linebegin = i * linewidth;
			memcpy(&pReadData[linebegin], &tmp[dwBitsSize - linewidth - linebegin], linewidth);
		}
		delete[] tmp;
	}
	//flip left <-> right
	if(sProps.btImgDescriptor & 16)
	{
		uint8_t *tmp = new uint8_t[dwBitsSize];
		if(not tmp)
		{
			delete[] pReadData;
			fprintf(stderr, "Error: out of memory\n");
			return 0;
		}

		//copy the image to temp memory
		memcpy(tmp, pReadData, dwBitsSize);

		//change pixels in every line
		for(uint32_t i = 0; i < sProps.wHeight; i++)
		{
			//copy new value to each pixel
			for(uint32_t j = 0; j < sProps.wWidth; j++)
			{
				//how many bytes there are in pixel
				uint32_t bytes = (uint32_t)(dwBitsSize / (sProps.wHeight * sProps.wWidth));
				uint32_t linewidth = dwBitsSize / sProps.wHeight;
				uint32_t pixel = j * bytes;
				uint32_t to = (i * linewidth) + pixel;
				uint32_t from = ((i + 1) * linewidth) - (pixel + bytes);
				memcpy(&pReadData[to], &tmp[from], bytes);
			}
		}
		delete[] tmp;
	}

	//image info
	size = dwBitsSize;
	width = sProps.wWidth;
	height = sProps.wHeight;
	//bits per pixel count equals the image size divided by pixel count,
	//multiplied by quantity of bits in byte (8)
	bpp = (size / (width * height)) * 8;
	return pReadData;
}

uint8_t *farReadImgData(char *szFarFileName, char *szFileName, ImgDims &dims)
{
	//open .far file
	FILE *pFARFile = fopen(szFarFileName, "rb");
	if(not pFARFile)
	{
		fprintf(stderr, "Error: cannot open file: \"%s\"\n", szFarFileName);
		return 0;
	}

	//check file size
	fseek(pFARFile, 0, SEEK_END);
	uint32_t dwFileSize = ftell(pFARFile);
	fseek(pFARFile, 0, SEEK_SET);

	//read .far header
	FarHeader sHeader;
	if(fread(&sHeader, 1, sizeof(FarHeader), pFARFile) < sizeof(FarHeader))
	{
		fprintf(stderr, "Error: cannot read file: \"%s\"\n", szFarFileName);
		fclose(pFARFile);
		return 0;
	}

	//check version, id and size
	if(sHeader.wFileVer != 100 or sHeader.wFileId != 0x141 or sHeader.dwFileSize != dwFileSize)
	{
		fprintf(stderr, "Error: file \"%s\" is invalid\n", szFarFileName);
		fclose(pFARFile);
		return 0;
	}

	//find image in .far
	FarRootEntry sRootEntry;
	int32_t i;
	for(i = 0; i < sHeader.iFilesQuantty; ++i)
	{
		//read root entry
		if(fread(&sRootEntry, 1, sizeof(FarRootEntry), pFARFile) < sizeof(FarRootEntry))
		{
			fprintf(stderr, "Error: cannot read file: \"%s\"\n", szFarFileName);
			fclose(pFARFile);
			return 0;
		}
		//read packed file name
		char *szPackedFileName = new char[sRootEntry.wFileNameLength];
		if(not szPackedFileName)
		{
			fprintf(stderr, "Error: out of memory while reading file: \"%s\"\n", szFarFileName);
			fclose(pFARFile);
			return 0;
		}
		if(fread(szPackedFileName, 1, sRootEntry.wFileNameLength * sizeof(char), pFARFile)
									< sRootEntry.wFileNameLength * sizeof(char))
		{
			delete[] szPackedFileName;
			fprintf(stderr, "Error: cannot read file: \"%s\"\n", szFarFileName);
			fclose(pFARFile);
			return 0;
		}
		//is this the file?
		if(0 == strcmp(szFileName, szPackedFileName))
		{
			delete[] szPackedFileName;
			break;
		}
		delete[] szPackedFileName;
	}
	//if we haven't found
	if(i == sHeader.iFilesQuantty)
	{
		fprintf(stderr, "Error: cannot find \"%s\" in \"%s\"\n", szFileName, szFarFileName);
		fclose(pFARFile);
		return 0;
	}

	//check if it is TGA file
	if(sRootEntry.wFileType != FAR_FMT_TGA)
	{
		fprintf(stderr, "Error: reading \"%s\"'s data from \"%s\" - "
						"format unsupported\n", szFileName, szFarFileName);
		fclose(pFARFile);
		return 0;
	}

	//set file pointer at the image file offset
	fseek(pFARFile, sRootEntry.dwPackedFileOffset, SEEK_SET);

	uint32_t dwWidth, dwHeight, dwBPP, dwSize;
	uint8_t *pData = readTGAData(pFARFile, dwWidth, dwHeight, dwBPP, dwSize);
	if(not pData)
	{
		fprintf(stderr, "Error: while reading \"%s\" in \"%s\"\n", szFileName, szFarFileName);
		fclose(pFARFile);
		return 0;
	}
	fclose(pFARFile);

	dims.dwWidth = dwWidth;
	dims.dwHeight = dwHeight;
	dims.dwSize = dwSize;
	dims.dwBPP = dwBPP;
	return pData;
}
